/****************************************************************************

 ■Software Radio
		Copyright(C) 2019 Denshi Usagi

****************************************************************************/

#include <xc.h>
#include <math.h>

//=================================================================

#define PI 3.141592653589793
typedef	int fractional;

#define PORT_LED	_LATB6

#define PLL_DIV		100	// FOSC:200MHz 100MIPS (160MIPS位までOK?)

#define TAP_NUM		16	// FIRタップ数(ADCが間に合うように12くらい以上)
#define OSC_NUM		8	// OSCデータ数(偶数)

// Sampling freq = 100 / 43 = 2.326MHz
// OSC ferq = 2.326 / 8 = 291KHz
// IQ freq = 164KHz, 746KHz

/*  --- FIR Filter Coeffs ---
	Remez Algorithm LPF
	Sampling Frequency = 2326.0
	cutoff1 = 180.0000000
	cutoff2 = 630.0000000
	Tap Count = 16
	attenuate = -90.00
	ripple factor = 1.0000000
	Quantized by 16 [bits]
*/
const float COEFFS[TAP_NUM] =
{
	-0.00134277344,
	-0.00674438477,
	-0.01504516602,
	-0.01376342773,
	 0.01910400391,
	 0.09536743164,
	 0.19424438477,
	 0.26553344727,
	 0.26553344727,
	 0.19424438477,
	 0.09536743164,
	 0.01910400391,
	-0.01376342773,
	-0.01504516602,
	-0.00674438477,
	-0.00134277344
};

#define KS_SIZE		((1 + TAP_NUM) * OSC_NUM * 2) // (sin + tap) * osc
#define DS_SIZE		(TAP_NUM * 2)	// delay 
#define PROC_SYCLE	((4 + TAP_NUM) * 2 + 3) // TAP=16の時43サイクル

//=================================================================

static fractional g_Ks[KS_SIZE] __attribute__((space(xmemory)));
static fractional g_Ds[DS_SIZE] __attribute__((space(ymemory)));
static fractional g_AD __attribute__((space(ymemory)));
static fractional* g_Dps[TAP_NUM];
static uint16_t* g_Dp;

//=================================================================

static void InitSystem(void);
static void InitApp(void);
static fractional ToQ15(float dVal);
static void Run(void);
static void LibDSP_FIR();

//=================================================================

// ●メインエントリ
int main(void)
{
	InitSystem();
	InitApp();
	Run();

	return 0;
}

// ●システム初期化
static void InitSystem(void)
{
	// クロック設定( FOSC = (4 * M) / (N1 * N2 * N3) )
	PLLFBDbits.PLLFBDIV = PLL_DIV; // M (400～1600, Default:150 max:200)
	CLKDIVbits.PLLPRE = 1;	// N1
	PLLDIVbits.POST1DIV = 2;// N2 (1～7, must POST1DIVx >= POST2DIVx def:4)
	PLLDIVbits.POST2DIV = 1;// N3 (1～7, must POST1DIVx >= POST2DIVx def:1)

	// ポート初期化
	PORTA = PORTB = 0x00;
	LATA = LATB = 0x00;

	// PORT_LED(RB6)
	TRISBbits.TRISB6 = 0;
	
	// DMA
	DMACONbits.DMAEN = 1; // Enables module
	DMAL = 0x1000;
	DMAH = 0x2000;

	// DMA CH0(ADCBUF0→g_AD)
	DMACH0bits.SAMODE = 0b00; // DMASRCn remains unchanged after a transfer completion
	DMACH0bits.DAMODE = 0b00; // DMADSTn remains unchanged after a transfer completion
	DMACH0bits.TRMODE = 0b01; // Repeated One-Shot
	DMACH0bits.RELOAD = 1; // DMASRC, DMADST and DMACNT registers are reloaded to their previous values upon the start of the next operation
	DMACH0bits.SIZE = 0; // Word (16-bit)
	DMAINT0bits.CHSEL = 0x28; // ADC Done AN0
	DMASRC0 = (uint16_t)&ADCBUF0;
	DMADST0 = (uint16_t)&g_AD;
	DMACNT0 = 1;
	DMACH0bits.CHEN = 1; // The corresponding channel is enabled
	
	// DMA CH1(g_Dps[n]→g_Dp)
	DMACH1bits.SAMODE = 0b01; // DMASRCn is incremented based on the SIZE bit after a transfer completion
	DMACH1bits.DAMODE = 0b00; // DMADSTn remains unchanged after a transfer completion
	DMACH1bits.TRMODE = 0b01; // Repeated One-Shot
	DMACH1bits.RELOAD = 1; // DMASRC, DMADST and DMACNT registers are reloaded to their previous values upon the start of the next operation
	DMACH1bits.SIZE = 0; // Word (16-bit)
	DMAINT1bits.CHSEL = 0x28; // ADC Done AN0
	DMASRC1 = (uint16_t)g_Dps;
	DMADST1 = (uint16_t)&g_Dp;
	DMACNT1 = TAP_NUM;
	DMACH1bits.CHEN = 1; // The corresponding channel is enabled
	
	// SCCP1(ADCのトリガー用にダブルエッジコンペアモードで)
	CCP1CON1Lbits.CLKSEL = 0b000; // FOSC/2(FP)
	CCP1CON1Lbits.TMRPS = 0b00; // 1:1 Prescaler
	CCP1CON1Lbits.T32 = 0; // Uses 16-bit time base for timer
	CCP1CON1Lbits.CCSEL = 0; // Output Compare/PWM/Timer peripheral
	CCP1CON1Lbits.MOD = 0b0100; // Dual Edge Compare mode
	CCP1CON1Hbits.TRIGEN = 0; // Trigger operation of time base is disabled
	CCP1CON1Hbits.SYNC = 0b00000; // None
	CCP1CON2Hbits.OCAEN = 0; // OCx pin is not controlled by the CCPx module
	CCP1PRL = PROC_SYCLE - 1;
	CCP1RBL = CCP1PRL;
	CCP1RAL = CCP1RBL - 1;
	CCP1TMRL = 0;

	// ADC(専用ADCコア0を使用)
	ADCON1Hbits.FORM = 1; // Fractional
	ADCON3Hbits.CLKSEL = 0b01; // FOSC
	ADCON3Hbits.CLKDIV = 0b00000; // 1 Source Clock Period
	ADCON4Hbits.C0CHS = 0b00; // S1AN0
	ADCORE0Hbits.RES = 0b11; // 12-bit resolution
	ADCORE0Hbits.ADCS = 0b0000000; // 2 Source Clock Periods
	ADTRIG0Lbits.TRGSRC0 = 0b10100; // Slave SCCP1 PWM/IC interrupt
	ADMOD0Lbits.DIFF0 = 0; // Channel is single-ended
	ADMOD0Lbits.SIGN0 = 1; // Channel output data is signed
	
	// ADC(専用ADCコア0)を有効化
	ADCON5Hbits.WARMTIME = 15; // 32768 Source Clock Periods
	ADCON1Lbits.ADON = 1; // ADC module is enabled
	ADCON5Lbits.C0PWR = 1; // ADC Core 0 is powered
	while(ADCON5Lbits.C0RDY == 0);
	ADCON3Hbits.C0EN = 1; // Dedicated ADC Core 0 is enabled

	// SCCP2(※テスト用 サイクル計測用にタイマモードで)
	CCP2CON1Lbits.CLKSEL = 0b000; // FOSC/2(FP)
	CCP2CON1Lbits.TMRPS = 0b00; // 1:1 Prescaler
	CCP2CON1Lbits.T32 = 0; // Uses 16-bit time base for timer
	CCP2CON1Lbits.CCSEL = 0; // Output Compare/PWM/Timer peripheral 
	CCP2CON1Lbits.MOD = 0b0000; // 16-Bit/32-Bit Timer mode
	CCP2CON1Hbits.TRIGEN = 0; // Trigger operation of time base is disabled
	CCP2CON1Hbits.SYNC = 0b00000; // None
	CCP2TMRL = 0;

	// ADC割り込み(DMAをトリガーできるようにする)
	ADIELbits.IE0 = 1;

	// AN0変換完了割り込み(※テスト用)
	//_ADCAN0IE = 1;
}

// ●アプリ初期化
static void InitApp(void)
{
	int i, j, k;

	// 係数テーブルの初期化
	for(i = 0, k = 0; i < OSC_NUM; i++)
	{
		// I(sin)
		g_Ks[k++] = ToQ15(sin(2 * PI * i / OSC_NUM));

		for(j = 0; j < TAP_NUM; j++)
			g_Ks[k++] = ToQ15(COEFFS[j]);

		// Q(cos)
		g_Ks[k++] = ToQ15(cos(2 * PI * i / OSC_NUM));

		for(j = 0; j < TAP_NUM; j++)
			g_Ks[k++] = ToQ15(COEFFS[j]);
	}

	// 遅延バッファの初期化
	for(i = 0; i < DS_SIZE; i++)
		g_Ds[i] = 0;

	// 遅延バッファ保存位置テーブルの初期化
	for(i = 0, j = DS_SIZE - 2; i < TAP_NUM; i++, j -= 2)
		g_Dps[i] = &g_Ds[j];
}

// ●double→Q15
static fractional ToQ15(float dVal)
{
	 float f = dVal * 32768.0;
	 
	 f = (f >= 0.0)? (f + 0.5) : (f - 0.5);
	 if (f >= 32767) return 0x7FFF;
	 if (f < -32768) return 0x8000;
	 
	 return (fractional)f;
}

// ●実行ループ
static void Run(void)
{
	PORT_LED = 1;

	// ※テスト用
	//LibDSP_FIR();
	
	// 飽和処理設定
	CORCONbits.SATA = 1; // Accumulator A saturation is enabled
	CORCONbits.SATB = 1; // Accumulator B saturation is enabled

	// モジュロアドレッシング設定
	XMODSRT = (unsigned)g_Ks;
	XMODEND = XMODSRT + sizeof(g_Ks) - 2;
	YMODSRT = (unsigned)g_Ds;
	YMODEND = YMODSRT + sizeof(g_Ds) - 2;
	MODCONbits.BWM = 0b1111; // Bit-reversed addressing disabled
	MODCONbits.XWM = 0b1000; // W8 selected for modulo addressing
	MODCONbits.YWM = 0b1010; // W10 selected for modulo addressing
	MODCONbits.XMODEN = 1; // X AGU modulus addressing enabled
	MODCONbits.YMODEN = 1; // Y AGU modulus addressing enabled

	// Master側がFIFO受信可能になるまで待つ
	while (SI1FIFOCSbits.SWFEN == 0); 

	// 変換開始！ADCのトリガータイマーON
	_DMA0IF = 0;
	CCP1CON1Lbits.CCPON = 1;
	//CCP2CON1Lbits.CCPON = 1; // サイクル計測用カウンタ(※テスト用)
	
	// 最初のDMA転送が完了するまで待つ(ADCの所要サイクルおよそ30サイクル)
	while(!_DMA0IF);

	// C変数をレジスタへ割り当てて初期化
	register fractional* pKs asm("w8")  = g_Ks;  // X Memory
	register fractional* pDs asm("w10") = g_Ds;  // Y Memory
	register fractional* pAD asm("w11") = &g_AD; // Y Memory
	(void)pKs; (void)pDs; (void)pAD;
	
	register uint16_t** nDp asm("w1")  = &g_Dp;
	register int nRept asm("w2") = TAP_NUM - 2 - 1;
	register volatile uint16_t* pFIFO asm("w3") = &SWMRFDATA;
	(void)nDp; (void)pFIFO; (void)nRept;
	
	// 電源断まで無限ループ
	asm("LOOP:");
	{
		asm("MOV [w1], w13");

		// I信号 ----------------------------------------------
		// 0 → A
		asm("CLR A, [w8]+=2, w4, [w11], w5");

		// g_Ks(sin[i]) * g_AD(ADC) → B
		asm("MPY w4 * w5, B, [w8]+=2, w4, [w10]+=4, w5");
		
		// A + g_Ks(coeffs[n]) * g_Ds(delay[n]) → A
		asm("REPEAT w2");
		asm("MAC w4 * w5, A, [w8]+=2, w4, [w10]+=4, w5");
		asm("MAC w4 * w5, A, [w8]+=2, w4, [w10]+=6, w5");
		asm("MAC w4 * w5, A, [W13]+= 2");
		
		// Round(A) → FIFO
		asm("SAC.R A, [w3]");

		// Q信号 ----------------------------------------------
		// 0 → A
		asm("CLR A, [w8]+=2, w4, [w11], w5");

		// g_Ks(cos[i]) * g_AD(ADC) → B
		asm("MPY w4 * w5, B, [w8]+=2, w4, [w10]+=4, w5");
		
		// A + g_Ks(coeffs[n]) * g_Ds(delay[n]) → A
		asm("REPEAT w2");
		asm("MAC w4 * w5, A, [w8]+=2, w4, [w10]+=4, w5");
		asm("MAC w4 * w5, A, [w8]+=2, w4, [w10]-=2, w5");
		asm("MAC w4 * w5, A, [W13]+= 2");

		// Round(A) → FIFO
		asm("SAC.R A, [w3]");
	}
	asm("BRA LOOP");
}

// ●AN0変換完了割り込み(テスト用)
void __attribute__((interrupt, shadow, no_auto_psv)) _ADCAN0Interrupt(void)
{
	static int nTime[100];
	static int nData[100];
	static int i = 0;

	int nVal = ADCBUF0;

	if (i >= 100)
	{
		// レベルチェック用
		static uint32_t c = 0;
		if (nVal < -((2048 - 205) << 4) || ((2048 - 205) << 4) < nVal)
		{
			// 入力過大の可能性がある
			PORT_LED = 0; 
			c = 500000;
		}
		else
		{
			// 範囲内
			if (c > 0) 
			{
				if (--c == 0)
					PORT_LED = 1;
			}
		}
	}
	else
	{
		nTime[i] = CCP2TMRL; // ADC周期確認用
		nData[i] = nVal;	 // AD値確認用
		i++;
	}

	_ADCAN0IF = 0;
}

//////////////////////////////////////////////////////////////////////
// テスト用
#include "dsp.h"

fractional Coeffs[TAP_NUM] __attribute__((space(xmemory), far));
fractional Delays[TAP_NUM] __attribute__((space(ymemory), far));
fractional fOut[TAP_NUM * 2] = { 0 };

// ●libdsp FIRフィルタ
static void LibDSP_FIR()
{
	int i;

	// 係数テーブルの初期化
	for(i = 0; i < TAP_NUM; i++)
		Coeffs[i] = ToQ15(COEFFS[i]);

	// FIRフィルタの初期化
	FIRStruct stFIR;
	FIRStructInit(&stFIR, TAP_NUM, Coeffs, COEFFS_IN_DATA, Delays);
	FIRDelayInit(&stFIR);
	
	// ダミーデータを処理させてみる
	for(i = 0; i < (TAP_NUM * 2); i++)
	{
		fractional fIn = 0x1234; // テスト入力値
		FIR(1, &fOut[i], &fIn, &stFIR);
	}
}
//////////////////////////////////////////////////////////////////////
